# Generated by Django 2.1.1 on 2018-09-21 00:10

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ('rc_django', '0003_auto_20170504_2136'),
    ]

    operations = [
        migrations.RenameField(
            model_name='cacheentry',
            old_name='header_pickle',
            new_name='headers',
        ),
    ]



"""
Contains memcached cache implementations
"""

import bmemcached
from bmemcached.exceptions import MemcachedException
import logging
import pickle
import threading
from dateutil.parser import parse
from django.conf import settings
from django.utils import timezone
from restclients_core.models import MockHTTP


logger = logging.getLogger(__name__)


class MemcachedCache(object):

    def __init__(self):
        self._set_client()

    def getCache(self, service, url, headers):
        key = self._get_key(service, url)
        try:
            data = self.client.get(key)
        except MemcachedException as ex:
            logger.error("Get (key: %s) ==> %s", key, str(ex))
            return

        if not data:
            return None

        values = pickle.loads(data)
        response = MockHTTP()
        response.headers = values["headers"]
        response.status = values["status"]
        response.data = values["data"]

        return {"response": response}

    def updateCache(self, service, url, new_data, new_data_dt):
        """
        :param new_data: a string representation of the data
        :param new_data_dt: a timezone aware datetime object giving
                the timestamp of the new_data
        :raise MemcachedException: if update failed
        """
        key = self._get_key(service, url)

        # clear existing data
        try:
            value = self.client.get(key)
            if value:
                data = pickle.loads(value)
                if "time_stamp" in data:
                    cached_data_dt = parse(data["time_stamp"])
                    if new_data_dt > cached_data_dt:
                        self.client.delete(key)
                        # may raise MemcachedException
                        logger.info("IN cache (key: %s), older DELETE", key)
                    else:
                        logger.info("IN cache (key: %s), newer KEEP", key)
                        return
            else:
                logger.info("NOT IN cache (key: %s)", key)

        except MemcachedException as ex:
            logger.error("Clear existing data (key: %s) ==> %s", key, str(ex))
            return

        # store new value in cache
        cdata, time_to_store = self._make_cache_data(
            service, url, new_data, {}, 200, new_data_dt)

        self.client.set(key, cdata, time=time_to_store)
        # may raise MemcachedException
        logger.info("MemCached SET (key %s) for %d seconds",
                    key, time_to_store)

    def _make_cache_data(self, service, url, data_to_cache,
                         header_data, status, time_stamp):
        data = {"status": status,
                "headers": header_data,
                "data": data_to_cache,
                "time_stamp": time_stamp.isoformat()
                }
        time_to_store = self.get_cache_expiration_time(service, url)
        return pickle.dumps(data), time_to_store

    def processResponse(self, service, url, response):
        header_data = {}
        for header in response.headers:
            header_data[header] = response.getheader(header)

        key = self._get_key(service, url)
        cdata, time_to_store = self._make_cache_data(service, url,
                                                     response.data,
                                                     header_data,
                                                     response.status,
                                                     timezone.now())
        try:
            self.client.set(key, cdata, time=time_to_store)
            logger.info("MemCached set with key '%s', %d seconds",
                        key, time_to_store)
        except bmemcached.exceptions.MemcachedException as ex:
            logger.error("set (key: %s) ==> %s", key, str(ex))
        return

    def get_cache_expiration_time(self, service, url):
        # Over-ride this to define your own.
        return 60 * 60 * 4

    def _get_key(self, service, url):
        return "%s-%s" % (service, url)

    def _set_client(self):
        thread_id = threading.current_thread().ident
        if not hasattr(MemcachedCache, "_memcached_cache"):
            MemcachedCache._memcached_cache = {}

        if thread_id in MemcachedCache._memcached_cache:
            self.client = MemcachedCache._memcached_cache[thread_id]
            return

        servers = settings.RESTCLIENTS_MEMCACHED_SERVERS
        username = getattr(settings, "RESTCLIENTS_MEMCACHED_USER", None)
        password = getattr(settings, "RESTCLIENTS_MEMCACHED_PASS", None)

        self.client = bmemcached.Client(servers, username, password)
        MemcachedCache._memcached_cache[thread_id] = self.client



from django.test import TestCase, RequestFactory
from django.test.utils import override_settings
from django.urls import reverse
from django.contrib.sessions.middleware import SessionMiddleware
from django.contrib.auth.middleware import AuthenticationMiddleware
from django.contrib.auth.models import User
from userservice.user import UserServiceMiddleware
from restclients_core.dao import DAO, MockDAO
from restclients_core.models import MockHTTP
from rc_django.views import errors
from rc_django.middleware import EnableServiceDegradationMiddleware
from rc_django.tests.test_views import missing_url
import time


class DELAY_DAO(DAO):
    def service_name(self):
        return "delay"

    def get_default_service_setting(self, key):
        if "DAO_CLASS" == key:
            return "rc_django.tests.performance.test_delay.Backend"


class Backend(MockDAO):
    def load(self, method, url, headers, body):
        response = MockHTTP()
        response.status = 200
        response.data = "ok"
        return response


class DegradedTestCase(TestCase):

    @override_settings(
        RESTCLIENTS_ADMIN_AUTH_MODULE='rc_django.tests.can_proxy_restclient')
    def test_degraded(self):
        r1 = RequestFactory().post(reverse("restclients_errors"), {
            "new_service_name": "delay",
            "new_service_status": 500,
            "new_service_content": "[oops",
            "new_service_load_time": 0.1,
        })
        r2 = RequestFactory().get("/")

        SessionMiddleware().process_request(r1)
        SessionMiddleware().process_request(r2)

        AuthenticationMiddleware().process_request(r1)
        UserServiceMiddleware().process_request(r1)

        AuthenticationMiddleware().process_request(r2)
        UserServiceMiddleware().process_request(r2)

        user = User.objects.create_user(username='delay_user',
                                        email='fake2@fake',
                                        password='top_secret')

        r1.user = user
        r2.user = user

        r1._dont_enforce_csrf_checks = True

        errors(r1)

        r1.session.save()

        EnableServiceDegradationMiddleware().process_request(r1)

        client = DELAY_DAO()
        t1 = time.time()
        response = client.getURL("/test", {})
        t2 = time.time()
        self.assertEquals(response.status, 500)
        self.assertEquals(response.data, "[oops")

        self.assertGreater(t2-t1, 0.09)

        EnableServiceDegradationMiddleware().process_request(r2)

        response = client.getURL("/test", {})
        self.assertEquals(response.status, 200)



import json
from unittest import skipIf
from django.conf import settings
from django.test import TestCase
from django.utils import timezone
from restclients_core.models import MockHTTP
from rc_django.cache_implementation.memcache import MemcachedCache


@skipIf(not getattr(settings, 'RESTCLIENTS_MEMCACHED_SERVERS', None),
        "Memcached cache not configured")
class MemcachedCacheTest(TestCase):
    def setUp(self):
        cache = MemcachedCache()
        cache.client.flush_all()

    def test_cacheGet(self):
        cache = MemcachedCache()
        key = cache._get_key('mem', '/same')
        self.assertEquals(key, "mem-/same")

        data = cache.getCache('mem', '/same', {})
        self.assertIsNone(data)

        cdata, time_to_store = cache._make_cache_data(
            'mem', '/same', json.dumps({"data": "Body Content"}),
            {}, 200, timezone.now())
        cache.client.set(key, cdata, time_to_store)

        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, '{"data": "Body Content"}')

    def test_updateCache(self):
        cache = MemcachedCache()
        # cache no data
        cache.updateCache('mem', '/same', '{"data": "Content1"}',
                          timezone.now())

        time1 = timezone.now()
        # cache has older data
        cache.updateCache('mem', '/same', '{"data": "Content2"}', time1)

        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.headers, {})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, '{"data": "Content2"}')

        # update with no newer data
        cache.updateCache('mem', '/same', '{"data": "Content3"}', time1)
        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.data, '{"data": "Content2"}')

    def test_processResponse(self):
        mock_resp = MockHTTP()
        mock_resp.status = 200
        mock_resp.data = "Content4"
        mock_resp.headers = {"Content-type": "text/html"}

        cache = MemcachedCache()
        cache.processResponse('mem', '/same1', mock_resp)

        hit = cache.getCache('mem', '/same1', {})
        response = hit["response"]
        self.assertEquals(response.headers, {"Content-type": "text/html"})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, "Content4")

    def test_binary_processResponse(self):
        mock_resp = MockHTTP()
        mock_resp.status = 200
        mock_resp.data = b'content to be encoded'
        mock_resp.headers = {"Content-type": "image/png"}

        cache = MemcachedCache()
        cache.processResponse('mem', '/same2', mock_resp)

        hit = cache.getCache('mem', '/same2', {})
        response = hit["response"]
        self.assertEquals(response.headers, {"Content-type": "image/png"})
        self.assertEquals(response.status, 200)
        self.assertEquals(response.data, b'content to be encoded')

    def test_memcached_client(self):
        cache = MemcachedCache()
        key = cache._get_key('mem', '/same')
        self.assertEquals(key, "mem-/same")
        cache.updateCache('mem', '/same', '{"data": "Content"}',
                          timezone.now())
        hit = cache.getCache('mem', '/same', {})
        response = hit["response"]
        self.assertEquals(response.headers, {})
        self.assertEquals(response.status, 200)
